@@ -7,7 +7,8 @@ if sys.version < '2.4':
     from sets import ImmutableSet as frozenset
 
 from routes import request_config
-from routes.util import controller_scan, RouteException
+from routes.lru import LRU
+from routes.util import controller_scan, MatchException, RoutesException
 from routes.route import Route
 
 
@@ -98,7 +99,7 @@ class Mapper(object):
         self.matchlist = []
         self.maxkeys = {}
         self.minkeys = {}
-        self.urlcache = {}
+        self.urlcache = LRU(4000)
         self._created_regs = False
         self._created_gens = False
         self.prefix = None
@@ -165,9 +166,12 @@ class Mapper(object):
             route.encoding = self.encoding
             route.decode_errors = self.decode_errors
         
-        self.matchlist.append(route)
+        if not route.static:
+            self.matchlist.append(route)
+        
         if routename:
             self._routenames[routename] = route
+            route.name = routename
         if route.static:
             return
         exists = False
@@ -263,7 +267,7 @@ class Mapper(object):
         if not self._created_regs and self.controller_scan:
             self.create_regs()
         elif not self._created_regs:
-            raise RouteException("You must generate the regular expressions"
+            raise RoutesException("You must generate the regular expressions"
                                  " before matching.")
         
         if self.always_scan:
@@ -301,7 +305,7 @@ class Mapper(object):
         
         """
         if not url:
-            raise RouteException('No URL provided, the minimum URL necessary'
+            raise RoutesException('No URL provided, the minimum URL necessary'
                                  ' to match is "/".')
         
         result = self._match(url)
@@ -345,7 +349,7 @@ class Mapper(object):
         
         if self.append_slash:
             kargs['_append_slash'] = True
-                
+        
         if not self.explicit:
             if 'controller' not in kargs:
                 kargs['controller'] = 'content'
@@ -373,14 +377,14 @@ class Mapper(object):
                 if key in self.urlcache:
                     return self.urlcache[key]
         
-        actionlist = self._gendict.get(controller) or self._gendict.get('*')
-        if not actionlist:
+        actionlist = self._gendict.get(controller) or self._gendict.get('*', {})
+        if not actionlist and not args:
             return None
         (keylist, sortcache) = actionlist.get(action) or \
-                               actionlist.get('*', (None, None))
-        if not keylist:
+                               actionlist.get('*', (None, {}))
+        if not keylist and not args:
             return None
-        
+
         keys = frozenset(kargs.keys())
         cacheset = False
         cachekey = unicode(keys)
@@ -454,8 +458,9 @@ class Mapper(object):
             if path:
                 if self.prefix:
                     path = self.prefix + path
+                external_static = route.static and route.external
                 if self.environ and self.environ.get('SCRIPT_NAME', '') != ''\
-                    and not route.absolute:
+                    and not route.absolute and not external_static:
                     path = self.environ['SCRIPT_NAME'] + path
                     key = cache_key_script_name
                 else:
@@ -675,22 +680,22 @@ class Mapper(object):
             for action in lst:
                 route_options['action'] = action
                 route_name = "%s%s_%s" % (name_prefix, action, collection_name)
-                self.connect(route_name, "%s/%s" % (collection_path, action),
-                                                    **route_options)
                 self.connect("formatted_" + route_name, "%s/%s.:(format)" % \
                              (collection_path, action), **route_options)
+                self.connect(route_name, "%s/%s" % (collection_path, action),
+                                                    **route_options)
             if primary:
                 route_options['action'] = primary
-                self.connect(collection_path, **route_options)
                 self.connect("%s.:(format)" % collection_path, **route_options)
+                self.connect(collection_path, **route_options)
         
         # Specifically add in the built-in 'index' collection method and its 
         # formatted version
-        self.connect(name_prefix + collection_name, collection_path, 
-                     action='index', conditions={'method':['GET']}, **options)
         self.connect("formatted_" + name_prefix + collection_name, 
             collection_path + ".:(format)", action='index', 
             conditions={'method':['GET']}, **options)
+        self.connect(name_prefix + collection_name, collection_path, 
+                     action='index', conditions={'method':['GET']}, **options)
         
         # Add the routes that deal with new resource methods
         for method, lst in new_methods.iteritems():
@@ -702,11 +707,11 @@ class Mapper(object):
                 if action != 'new':
                     name = action + "_" + name
                 route_options['action'] = action
-                self.connect(name_prefix + name, path, **route_options)
-                path = (action == 'new' and new_path + '.:(format)') or \
+                formatted_path = (action == 'new' and new_path + '.:(format)') or \
                     "%s/%s.:(format)" % (new_path, action)
-                self.connect("formatted_" + name_prefix + name, path, 
+                self.connect("formatted_" + name_prefix + name, formatted_path, 
                              **route_options)
+                self.connect(name_prefix + name, path, **route_options)
         
         requirements_regexp = '[^\/]+'
 
@@ -720,19 +725,71 @@ class Mapper(object):
                 primary = None
             for action in lst:
                 route_options['action'] = action
-                self.connect("%s%s_%s" % (name_prefix, action, member_name),
-                    "%s/%s" % (member_path, action), **route_options)
                 self.connect("formatted_%s%s_%s" % (name_prefix, action, 
                                                     member_name),
                     "%s/%s.:(format)" % (member_path, action), **route_options)
+                self.connect("%s%s_%s" % (name_prefix, action, member_name),
+                    "%s/%s" % (member_path, action), **route_options)
             if primary:
                 route_options['action'] = primary
+                self.connect("%s.:(format)" % member_path, **route_options)
                 self.connect(member_path, **route_options)
         
         # Specifically add the member 'show' method
         route_options = requirements_for('GET')
         route_options['action'] = 'show'
         route_options['requirements'] = {'id':requirements_regexp}
-        self.connect(name_prefix + member_name, member_path, **route_options)
         self.connect("formatted_" + name_prefix + member_name, 
                      member_path + ".:(format)", **route_options)
+        self.connect(name_prefix + member_name, member_path, **route_options)
+    
+    def redirect(self, match_path, destination_path, *args, **kwargs):
+        """Add a redirect route to the mapper
+        
+        Redirect routes bypass the wrapped WSGI application and instead
+        result in a redirect being issued by the RoutesMiddleware. As
+        such, this method is only meaningful when using
+        RoutesMiddleware.
+        
+        By default, a 302 Found status code is used, this can be
+        changed by providing a ``_redirect_code`` keyword argument
+        which will then be used instead. Note that the entire status
+        code string needs to be present.
+        
+        When using keyword arguments, all arguments that apply to
+        matching will be used for the match, while generation specific
+        options will be used during generation. Thus all options
+        normally available to connected Routes may be used with
+        redirect routes as well.
+        
+        Example::
+            
+            map = Mapper()
+            map.redirect('/legacyapp/archives/{url:.*}, '/archives/{url})
+            map.redirect('/home/index', '/', _redirect_code='301 Moved Permanently')
+        
+        """
+        both_args = ['_encoding', '_explicit', '_minimize']
+        gen_args = ['_filter']
+        
+        status_code = kwargs.pop('_redirect_code', '302 Found')
+        gen_dict, match_dict = {}, {}
+        
+        # Create the dict of args for the generation route
+        for key in both_args + gen_args:
+            if key in kwargs:
+                gen_dict = kwargs[key]
+        gen_dict['_static'] = True
+        
+        # Create the dict of args for the matching route
+        for key in kwargs:
+            if key not in gen_args:
+                match_dict[key] = kwargs[key]
+        
+        self.connect(match_path, **match_dict)
+        match_route = self.matchlist[-1]
+        
+        self.connect('_redirect_%s' % id(match_route), destination_path,
+                     **gen_dict)
+        match_route.redirect = True
+        match_route.redirect_status = status_code
